#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <cmath>
#include <iomanip>
#include <ctime>
#include "finum.hpp"

/* One-dimensional Normal Law. Density function */

double nd(double x)
{
  return(exp(-SQR(x)/2.)/(sqrt(2.*PI)));
}

/*One-Dimensional Normal Law. Cumulative distribution function. */
/*Abramowitz, Milton et Stegun, Handbook of MathematicalFunctions, 1968, Dover Publication, New York, page 932 (26.2.18).Precision 10-7*/
double N(double x)
{
  const double p= 0.2316419;
  const double b1= 0.319381530;
  const double b2= -0.356563782;
  const double b3= 1.781477937;
  const double b4= -1.821255978;
  const double b5= 1.330274429;
  const double one_over_twopi= 0.39894228;
  
  double t;
  
  if(x >= 0.0)
    {
      t = 1.0 / ( 1.0 + p * x );
      return (1.0 - one_over_twopi * exp( -x * x / 2.0 ) * t * ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
    } 
  else 
    {/* x < 0 */
      t = 1.0 / ( 1.0 - p * x );
      return ( one_over_twopi * exp( -x * x / 2.0 ) * t * ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
    }
}




/*-------------------------------------------------------------------*/

double Inverse_erf_Moro(double x)    
{
  static double a[4] = {
    2.50662823884, 
    -18.61500062529, 
    41.39119773534, 
    -25.44106049636};
  static double b[4] = {
    -8.47351093090,
    23.08336743743 , 
    -21.06224101826, 
    3.13082909833};
  static double c[9] ={
    0.337475482276147,
    0.9761690190917186,
    0.1607979714918209,
    0.0276438810333863,
    0.0038405729373609,
    0.0003951896511919,
    0.0000321767881768,
    0.000002888167364,
    0.0000003960315187};
  
  double u,r;
  
  u=x-0.5;
  if (fabs(u)<0.42) 
    {
      r=u*u;
      r=u*(((a[3]*r+a[2])*r+a[1])*r+a[0])/((((b[3]*r+b[2])*r+b[1])*r+b[0])*r+1.0);
      return(r);
    }

  r=x;
  if(u>0.0) 
    r=1.0-x;
  r=log(-log(r));
  r=c[0]+r*(c[1]+r*(c[2]+r*(c[3]+r*(c[4]+r*(c[5]+r*c[6]+r*(c[7]+r*c[8]))))));
  if (u<0.0) 
    r=-r;
  return(r);
}



/*-------------------------------------------------------------------*/
#define TailleTable 32 
 
/* pour le generateur direct */  
#define m1 2147483563 
#define m2 2147483399 
#define a1 40014 
#define a2 40692 
#define qa1 53668 
#define qa2 52774 
#define ra1 12211 
#define ra2 3791 
#define hom (1+(m1-1)/TailleTable) 
 


#define MAXBIT 30
#define MAXDIM 160


void tsb2_seq(int *n, float x[])
{
    int                     j, k, l;
    unsigned long           i, im, ipp;

    static unsigned long    in;
    static float            fac;
    static unsigned long    ix[MAXDIM + 1];
    static unsigned long    *iu[MAXBIT + 1];
    /*
        static unsigned long mdeg[MAXDIM+1]={0,1,2,3,3,4,4};
        static unsigned long ip[MAXDIM+1]={0,0,1,1,2,1,4};
        static unsigned long iv[MAXDIM*MAXBIT+1]={
                0,1,1,1,1,1,1,3,1,3,3,1,1,5,7,7,3,3,5,15,11,5,15,13,9};
	*/

        static unsigned long  mdeg[MAXDIM+1] = {0 ,1 ,2 ,3 ,3 ,4 ,4 ,5 ,5 ,5 ,5 ,5 ,5 ,6 ,6 ,6 ,6 ,6 ,6 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,8 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,9 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10};
 
        static unsigned long  ip[MAXDIM+1] = {0 ,0 ,1 ,1 ,2 ,1 ,4 , 2 ,4 ,7 ,11 ,13 ,14 ,1 ,13 ,16 ,19 ,22 ,25 ,1 ,4 ,7 ,8 ,14 ,19 ,21 ,28 ,31 ,32 ,37 ,41 ,42 ,50 ,55 ,56 ,59 ,62 ,14 ,21 ,22 ,38 ,47 ,49 ,50 ,52 ,56 ,67 ,70 ,84 ,97 ,103 ,115 ,122 ,8 ,13 ,16 ,22 ,25 ,44 ,47 ,52 ,55 ,59 ,62 ,67 ,74 ,81 ,82 ,87 ,91 ,94 ,103 ,104 ,109 ,122 ,124 ,137 ,138 ,143 ,145 ,152 ,157 ,167 ,173 ,176 ,181 ,182 ,185 ,191 ,194 ,199 ,218 ,220 ,227 ,229 ,230 ,234 ,236 ,241 ,244 ,253 ,4 ,13 ,19 ,22 ,50 ,55 ,64 ,69 ,98 ,107 ,115 ,121 ,127 ,134 ,140 ,145 ,152 ,158 ,161 ,171 ,181 ,194 ,199 ,203 ,208 ,227 ,242 ,251 ,253 ,265 ,266 ,274 ,283 ,289 ,295 ,301 ,316 ,319 ,324 ,346 ,352 ,361 ,367 ,382 ,395 ,398 ,400 ,412 ,419 ,422 ,426 ,428 ,433 ,446 ,454 ,457 ,472 ,493 ,505 ,508};
 
 
        static unsigned long  iv[MAXDIM*MAXBIT+1]={0 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,3 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,1 ,5 ,7 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,3 ,15 ,11 ,13 ,15 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,7 ,17 ,13 ,5 ,23 ,25 ,31 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,15 ,51 ,61 ,47 ,53 ,9 ,47 ,57 ,61 ,53 ,55 ,51 ,59 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,31 ,85 ,67 ,113 ,29 ,11 ,109 ,25 ,29 ,21 ,71 ,75 ,83 ,97, 101 ,127 ,111 ,119 ,99 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,63 ,255 ,79 ,145 ,149 ,191 ,173 ,27 ,151 ,191 ,229 ,73 ,81 ,33 ,37 ,191 ,143 ,167 ,155 ,193 ,241 ,209 ,249 ,233 ,221 ,205 ,245 ,213 ,255 ,207 ,199 ,231 ,227 ,211 ,251 ,219 ,235 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,127 ,257 ,465 ,147 ,183 ,449 ,181 ,303 ,147 ,491 ,101 ,425 ,209 ,35 ,335 ,445 ,461 ,389 ,153 ,65 ,113 ,81 ,121 ,105 ,93 ,77 ,117 ,85 ,383 ,303 ,311 ,343 ,347 ,267 ,371 ,275 ,323 ,465 ,409 ,473 ,477 ,429 ,389 ,453 ,485 ,501 ,447 ,479 ,487 ,387 ,419 ,443 ,459 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,255 ,771 ,721 ,445 ,411 ,193 ,949 ,295 ,947 ,235 ,109 ,169 ,913 ,39 ,923 ,701 ,589 ,133 ,889 ,67 ,115 ,83 ,123 ,107 ,727 ,647 ,607 ,767 ,893 ,1005 ,981 ,821 ,345 ,265 ,369 ,273 ,321 ,209 ,153 ,217 ,221 ,173 ,133 ,197 ,229 ,245 ,575 ,671 ,727 ,635 ,539 ,563 ,675 ,993 ,801 ,1009 ,945 ,785 ,985 ,857 ,969 ,841 ,873 ,937 ,893 ,925 ,781 ,909 ,845 ,877 ,941 ,837 ,997 ,805 ,917 ,981 ,799 ,927 ,863 ,783 ,1007 ,815 ,839 ,807 ,1015 ,823 ,951 ,791 ,855 ,899 ,835 ,915 ,979 ,891 ,827 ,955 ,923 ,987 ,779 ,971 ,811 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511 ,511};
 

    if (*n < 0) 
    {
       for (k = 1; k <= MAXDIM; k++)  ix[k] = 0;

       in=0;

       if (iv[1] != 1) return;

       fac=1.0/(1L << MAXBIT);

       for (j = 1, k = 0; j <= MAXBIT; j++, k += MAXDIM) 
          iu[j] = &iv[k];

       for (k = 1; k <= MAXDIM; k++) 
       {
          for (j = 1; j <= mdeg[k]; j++) 
             iu[j][k] <<= (MAXBIT - j);

          for (j = mdeg[k] + 1; j <= MAXBIT; j++) 
          {
             ipp = ip[k];

             i = iu[j-mdeg[k]][k];

             i ^= (i >> mdeg[k]);

             for (l = mdeg[k] - 1; l >= 1; l--) 
             {
                if (ipp & 1)  i ^= iu[j-l][k];

                ipp >>= 1;
             }

             iu[j][k]=i;
          }

       }

    } 
    else 
    {
       im=in;

       for (j = 1; j <= MAXBIT; j++) 
       {
          if (!(im & 1)) break;

          im >>= 1;
       }

       if (j > MAXBIT) return;

       im=(j-1)*MAXDIM;

       for (k = 1; k <= MIN(*n, MAXDIM); k++) 
       {
          ix[k] ^= iv[im+k];

          x[k]=ix[k]*fac;
       }

       in++;
    }


}

// Closed formulae for the BlackScholes Call price and delta.
// return OK
int Call_BlackScholes_73(double s, double Maturity, double Strike, double r,double divid, double sigma, double *ptprice, double *ptdelta, double *ptgamma)
{
  double sigmasqrt,d1,d2,delta;
  double t,k;

  t=Maturity;
  k=Strike;

  if(k>0)
    {
  sigmasqrt=sigma*sqrt(t);
  d1=(log(s/k)+(r-divid)*t)/sigmasqrt+sigmasqrt*0.5;
  d2=d1-sigmasqrt;
  delta=exp(-divid*t)*N(d1);

  /*Price*/

  *ptprice=s*delta-exp(-r*t)*k*N(d2);

	
  /*Delta*/
  *ptdelta=delta;

  /*Gamma*/
  *ptgamma = exp(-0.5*d1*d1 - divid*t)/sqrt(2.*PI)/sigmasqrt/s;
     }
  else
    {
      *ptprice = s*exp(-divid*t);
      *ptdelta = 1;
      *ptgamma = 0.;
    }
  return 0;
}

// Closed formulae for the BlackScholes Put price and delta.
// return OK
int Put_BlackScholes_73(double s, double Maturity, double Strike, double r,double divid, double sigma, double *ptprice, double *ptdelta, double *ptgamma)
{
  double sigmasqrt,d1,d2,delta;
  double t,k;

  t=Maturity;
  k=Strike;

  if(k>0)
    {
  sigmasqrt=sigma*sqrt(t);
  d1=(log(s/k)+(r-divid)*t)/sigmasqrt+sigmasqrt*0.5;
  d2=d1-sigmasqrt;
  delta=-exp(-divid*t)*N(-d1);

  /*Price*/

  *ptprice=exp(-r*t)*k*N(-d2)+delta*s;

	
  /*Delta*/
  *ptdelta=delta;

  /*Gamma*/
  *ptgamma = exp(-0.5*d1*d1 - divid*t)/sqrt(2.*PI)/sigmasqrt/s;
     }
  else
    {
      *ptprice = 0.;
      *ptdelta = -1;
      *ptgamma = 0.;
    }
  return 0;
}

int ImpliedVol(double price,double s0,double Maturity, double Strike,double eqr, double eqdiv,double sigma_ref,double *volimpl, double *impliedDelta, double *impliedGamma)
{

  double result1,minor,major,price_major,price_minor,eps=PRICE_ACCURACY, dummy, dummy1, K,sigma;

  minor=major=sigma_ref;

  Put_BlackScholes_73(s0, Maturity, Strike, eqr, eqdiv, sigma_ref, &result1, &dummy, &dummy1);

  K=Strike;



  // Dichotomy to find the implied volatility. If the length of the interval (minor, major) is < eps*FAC_SLOPE, the Dichotomy is breaked.
  if (result1 > price)
    {
      do {
        minor=minor*0.5;
        Put_BlackScholes_73(s0, Maturity, Strike, eqr, eqdiv, minor, &result1, &dummy, &dummy1);

        //cout << "sigma:"<<minor<<"   prix:"<<result1<<"   prixm:"<<Price<<"   intr:"<< Payoff(s0,Strike) <<"  bound:"<<s0-exp(-eqr*Maturity)*Strike <<endl;
        //getchar();

        //if (minor<0.0001) return -1;

      } while( result1 >price && fabs(result1-price)> eps);
      price_minor=result1;
    }
  else
    {
      do {
        major=major*2;
        Put_BlackScholes_73(s0, Maturity, Strike, eqr, eqdiv, major, &result1, &dummy, &dummy1);

      } while( result1 <price && fabs(result1-price)> eps);
      price_major=result1;
    }

  sigma=(minor+major)*0.5;

  Put_BlackScholes_73(s0, Maturity, Strike, eqr, eqdiv, sigma, &result1, &dummy, &dummy1);

  while((fabs(result1- price)>eps)&&(major-minor>eps*FAC_SLOPE))
    {
      if(result1>price)
        {
          major=sigma;
        }
      else
        {
          minor=sigma;
        }

      sigma=(minor+major)*0.5;
      Put_BlackScholes_73(s0, Maturity, Strike, eqr, eqdiv, sigma, &result1, &dummy, &dummy1);


    }
  *volimpl= sigma;
  *impliedDelta = dummy;
  *impliedGamma = dummy1;

  return 0;
}


void Progonka_in_place(const  std::vector<double>& low, 
					   const  std::vector<double>& diag, const  std::vector<double>& up,  std::vector<double>& rightside_result)
{
      int N = diag.size();
      if((low.size()!=N-1)||(low.size()!=N-1)||(rightside_result.size()!=N)) {
		  std::cout << "les tailles des vecteurs sont incorrectes!" << std::endl;
		  std::exit(1);
      }

      std::vector<double> A(N-1), B(N-1);

      // descente
      const std::vector<double> & rightside = rightside_result;
      A[0] = -up[0]/diag[0];
      B[0] = rightside[0]/diag[0];
      double denom;
      for(int i=1;i<N-1;i++)
      {
        denom = diag[i] + low[i-1]*A[i-1];
        A[i] = -up[i]/denom;
        B[i] = (rightside[i]-low[i-1]*B[i-1])/denom;
      };

      // remontee
      std::vector<double> & result = rightside_result;
      result[N-1] = (rightside[N-1] - low[N-2]*B[N-2])/(diag[N-1] + low[N-2]*A[N-2]);
      for(int i=N-2;i>=0;i--)
      {
        result[i] = A[i]*result[i+1] + B[i];
      };
}



void Progonka( std::vector<double>::const_iterator low_begin,  
			  std::vector<double>::const_iterator diag_begin,  std::vector<double>::const_iterator up_begin, 
			std::vector<double>::const_iterator rs_begin,  std::vector<double>::iterator result_begin,  std::vector<double>::iterator result_end)
{
      int N = result_end - result_begin;

      std::vector<double> A(N-1), B(N-1);

      // descente
       std::vector<double>::const_iterator low_iter = low_begin;
       std::vector<double>::const_iterator diag_iter = diag_begin;
       std::vector<double>::const_iterator up_iter = up_begin;
       std::vector<double>::const_iterator rs_iter = rs_begin;


      A[0] = -(*up_iter)/(*diag_iter);
      B[0] = (*rs_iter)/(*diag_iter);
      double denom;
      for(int i=1;i<N-1;i++)
      {
        diag_iter++;
        up_iter++;
        rs_iter++;

        denom = (*diag_iter) + (*low_iter)*A[i-1];
        A[i] = -(*up_iter)/denom;
        B[i] = ((*rs_iter) - (*low_iter)*B[i-1])/denom;

        low_iter++;
      };


      // remontee
       std::vector<double>::iterator result_iter = --result_end;
      diag_iter++;
      rs_iter++;
      *result_iter = ((*rs_iter) - (*low_iter)*B[N-2])/((*diag_iter) + (*low_iter)*A[N-2]);
      result_iter--;

      for(int i = N - 2; i >= 0; i--)
      {
         *(result_iter) = A[i]*(*(result_iter+1)) + B[i];
         result_iter--;
      };
  }


///================== P R I C E === N O N E ==============================================================================

int TreePriceNone(double s0, Vanilla Type, double Maturity,  double Strike, double r,double  divid, double sigma, double *ptprice, double *ptdelta, double *ptgamma, int nTime, double lambda)
{ 
  return 0;
}


int  ImplPriceNone(double s0, Vanilla Type, double Maturity,  double Strike, double r,double  divid, double sigma, double *ptprice, double *ptdelta, double *ptgamma, int nTime,int nSpace)
{ 
  return 0;
}


int  CNPriceNone(double s0, Vanilla Type, double Maturity,  double Strike, double r,double  divid, double sigma, double *ptprice, double *ptdelta, double *ptgamma, int nTime,int nSpace)
{ 
  return 0;
}

///===================== U P === A N D === O U T ============================================================
int TreePriceUpOut(double s0,  Vanilla Type, double Maturity,  double Strike, double barrier,  double rebate, double r,double  divid, double sigma, double *ptprice, double *ptdelta, double *ptgamma, int nTime, double lambda)
{ 
  return 0;
}


int  ImplPriceUpOut(double s0,  Vanilla Type, double Maturity,  double Strike, double barrier,  double rebate, double r,double  divid, double sigma, double *ptprice, double *ptdelta, double *ptgamma, int nTime,int nSpace)
{ 
  return 0;
}


using namespace std;

int main(int argc, char* argv[]) {
  cout << Inverse_erf_Moro(0.5)<< endl;

  /////////
  std::ofstream outfile("sortie.dat");
  float tsb2[MAXDIM+1];
  int N=5;
  //initialization of tsb2
  int m=-1;
  tsb2_seq(&m, tsb2);
  m=MAXDIM;

  for(int i=0;i<N;i++)//draw N quasi-uniform point in [0;1]^m
    {
      tsb2_seq(&m,tsb2);//tsb2[1..m]=quasi-uniform point in [0;1]^m
	  outfile <<  tsb2[1] << " " << tsb2[2] << std::endl;
    }
  ///////
  double price,s0, Maturity,  Strike, eqr,  eqdiv, sigma_ref, volimpl,impliedDelta,  impliedGamma;
  price=.2;
  s0=1.;
  Maturity=1.;
  Strike=1.;
  eqr=0.;
  eqdiv=0.;
  sigma_ref=.2;
  volimpl=0.3;

  ImpliedVol(price, s0, Maturity,  Strike, eqr,  eqdiv, sigma_ref, &volimpl,  &impliedDelta,  &impliedGamma);

  Put_BlackScholes_73(s0,  Maturity,  Strike,  eqr,  eqdiv,  volimpl, &price, &impliedDelta,  &impliedGamma) ;

  cout << volimpl << " " << price << " " <<  impliedDelta << " " << impliedGamma << " " << endl ;

  /////
 int nSpace = 3;

  std::vector<double> Price(nSpace + 1);

  // ces vecteurs representent la matrice tridiagonale du systeme:
  std::vector<double> low(nSpace);      // diagonale inferieure
  std::vector<double> diag(nSpace + 1); // diagonale principale
  std::vector<double> up(nSpace);       // diagonale superieure


  // conditions au bord: propagation des valeurs initiales:
  up[0] = 0;
  diag[0] = 1;

  low[nSpace - 1] = 0;
  diag[nSpace] = 1;

  for(int i = 1; i < nSpace; i++)
    {
      Price[i] = 0;
    }

  for(int i = 1; i < nSpace; i++)
    {
      up[i] = 0;
      diag[i] = 1;
      low[i-1] = 0;
    }

  Progonka_in_place(low, diag, up, Price);

  for(int i = 1; i < nSpace; i++)
    {
		cout << i << " "<< up[i]<< " " << diag[i]<< " " << low[i-1]  << " " <<  
		Price[i]  << endl;
    }

  return 0;
}


